/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov,  2008, 2011, 2014
*
****************************************************************************/
package inversion_applet;

import inversion_applet.DataCenter;
import inversion_applet.Utilities;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.GridLayout;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JTextField;
import javax.swing.JPanel;
import javax.swing.JOptionPane;
import javax.swing.SwingConstants;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.File;

public class ConstantsPage extends JPanel {
  private DataCenter dataCenter;
  private JLabel GravitationalLabel;
  private JLabel GravitationalCurrentValue;
  private JPanel GravitationalPanel;
  private JTextField GravitationalTextField;
  private JButton GravitationalUpdate;
  private JButton GravitationalButton1;
  private JButton GravitationalButton2;
  private JButton GravitationalButton3;
  private JButton GravitationalButton4;
  private JButton GravitationalButton5;
  private JLabel SolarMassLabel;
  private JLabel SolarMassCurrentValue;
  private JPanel SolarMassPanel;
  private JTextField SolarMassTextField;
  private JButton SolarMassUpdate;
  private JButton SolarMassButton1;
  private JButton SolarMassButton2;
  private JButton SolarMassButton3;
  private JLabel SolarRadiusLabel;
  private JLabel SolarRadiusCurrentValue;
  private JPanel SolarRadiusPanel;
  private JTextField SolarRadiusTextField;
  private JButton SolarRadiusUpdate;
  private JButton SolarRadiusButton1;
  private JButton SolarRadiusButton2;
  private JButton SaveConstantsButton;

  /////////////////////////////////////////////////////////////////////////////
  /** A constructor */
  /////////////////////////////////////////////////////////////////////////////
  public ConstantsPage(DataCenter dataCenter) {
    this.dataCenter = dataCenter;
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This is what sets up the buttons and the layout */
  /////////////////////////////////////////////////////////////////////////////
  public void init() {
    initGravitationalButtons();
    initSolarMassButtons();
    initSolarRadiusButtons();
    initSaveConstantsButton();

    setLayout(new GridLayout(23,1));
    add(GravitationalLabel);
    add(GravitationalCurrentValue);
    add(GravitationalPanel);
    add(GravitationalButton1);
    add(GravitationalButton2);
    add(GravitationalButton3);
    add(GravitationalButton4);
    add(GravitationalButton5);
    add(new JLabel(""));
    add(SolarMassLabel);
    add(SolarMassCurrentValue);
    add(SolarMassPanel);
    add(SolarMassButton1);
    add(SolarMassButton2);
    add(SolarMassButton3);
    add(new JLabel(""));
    add(SolarRadiusLabel);
    add(SolarRadiusCurrentValue);
    add(SolarRadiusPanel);
    add(SolarRadiusButton1);
    add(SolarRadiusButton2);
    add(new JLabel(""));
    add(SaveConstantsButton);
  }

  ////////////////////////////////////////////////////////////////////////////////////
  // This initialises all the buttons used for setting the graviational constant
  // with different values found in the literature.
  ////////////////////////////////////////////////////////////////////////////////////
  private void initGravitationalButtons() {
    GravitationalLabel = new JLabel("<html><h2><font color='blue'>Gravitational constant, G</font></h2></html>");
    GravitationalCurrentValue = new JLabel(String.format("Current value: %.15e (cm\u00B3.g\u207B\u00B9.s\u207B\u00B2)",dataCenter.G));
    dataCenter.GravitationalCurrentValue = GravitationalCurrentValue;
    GravitationalTextField = new JTextField(String.format("%.15e ",dataCenter.G));
    GravitationalTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        Double G = Utilities.readDouble(GravitationalTextField.getText(),
                     "gravitational constant",5e-8,8e-8,GravitationalTextField);
        if (G == null) {
          GravitationalTextField.setText(String.format("%.15e",dataCenter.G));
        } else {
          GravitationalTextField.setText(String.format("%.15e",G));
        }
      }
    });

    GravitationalUpdate = new JButton("Update value");
    GravitationalUpdate.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        Double G = Utilities.readDouble(GravitationalTextField.getText(),
                     "gravitational constant",5e-8,8e-8,GravitationalUpdate);
        if (G != null) {
          dataCenter.modifyG(G.doubleValue());  // if G is modified, the model also has to be updated
          GravitationalCurrentValue.setText(String.format("Current value: %.15e (cm\u00B3.g\u207B\u00B9.s\u207B\u00B2)",dataCenter.G));
        }
        GravitationalTextField.setText(String.format("%.15e",dataCenter.G));
      }
    });

    GravitationalPanel = new JPanel(new GridLayout(1,2));
    GravitationalPanel.add(GravitationalTextField);
    GravitationalPanel.add(GravitationalUpdate);

    GravitationalButton1 = new JButton("G = 6.6716823\u00D710\u207B\u2078 (GraCo, OSCROX, POSC, PULSE, CoRoT/ESTA models)");
    GravitationalButton1.addActionListener(GravitationalListener(6.6716823e-8));
    GravitationalButton1.setHorizontalAlignment(SwingConstants.LEFT);
    GravitationalButton2 = new JButton("G = 6.67232\u00D710\u207B\u2078 (ADIPLS, FILOU, LOSC, LNAWENR, Model S)");
    GravitationalButton2.addActionListener(GravitationalListener(6.67232e-8));
    GravitationalButton2.setHorizontalAlignment(SwingConstants.LEFT);
    GravitationalButton3 = new JButton("G = 6.67259\u00D710\u207B\u2078 (NOSC)");
    GravitationalButton3.addActionListener(GravitationalListener(6.67259e-8));
    GravitationalButton3.setHorizontalAlignment(SwingConstants.LEFT);
    GravitationalButton4 = new JButton("G = 6.67384\u00D710\u207B\u2078 (2010 CODATA)");
    GravitationalButton4.addActionListener(GravitationalListener(6.67384e-8));
    GravitationalButton4.setHorizontalAlignment(SwingConstants.LEFT);
    GravitationalButton5 = new JButton("G = 6.67428\u00D710\u207B\u2078 (MESA)");
    GravitationalButton5.addActionListener(GravitationalListener(6.67428e-8));
    GravitationalButton5.setHorizontalAlignment(SwingConstants.LEFT);
  }

  ////////////////////////////////////////////////////////////////////////////////////
  // This generates action listeners for the Gravitational buttons.
  ////////////////////////////////////////////////////////////////////////////////////
  private ActionListener GravitationalListener(final double value) {
    return new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        dataCenter.modifyG(value);  // if G is modified, the model also has to be updated
        GravitationalCurrentValue.setText(String.format("Current value: %.15e (cm\u00B3.g\u207B\u00B9.s\u207B\u00B2)",dataCenter.G));
        GravitationalTextField.setText(String.format("%.15e",dataCenter.G));
      }
    };
  }

  ////////////////////////////////////////////////////////////////////////////////////
  // This initialises all the buttons used for setting the graviational constant
  // with different values found in the literature.
  ////////////////////////////////////////////////////////////////////////////////////
  private void initSolarMassButtons() {
    SolarMassLabel = new JLabel("<html><h2><font color='blue'>Solar mass, M<sub>\u2609</sub></font></h2></html>");
    SolarMassCurrentValue = new JLabel(String.format("Current value: %.15e (g)",dataCenter.SolarMass));
    SolarMassTextField = new JTextField(String.format("%.15e",dataCenter.SolarMass));
    SolarMassTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        Double SolarMass = Utilities.readDouble(SolarMassTextField.getText(),
                     "solar mass",1e33,3e33,SolarMassTextField);
        if (SolarMass == null) {
          SolarMassTextField.setText(String.format("%.15e ",dataCenter.SolarMass));
        } else {
          SolarMassTextField.setText(String.format("%.15e ",SolarMass));
        }
      }
    });

    SolarMassUpdate = new JButton("Update value");
    SolarMassUpdate.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        Double SolarMass = Utilities.readDouble(SolarMassTextField.getText(),
                     "solar mass",1e33,3e33,SolarMassUpdate);
        if (SolarMass != null) {
          dataCenter.SolarMass = SolarMass;
          SolarMassCurrentValue.setText(String.format("Current value: %.15e (g)",dataCenter.SolarMass));
        }
        SolarMassTextField.setText(String.format("%.15e ",dataCenter.SolarMass));
      }
    });

    SolarMassPanel = new JPanel(new GridLayout(1,2));
    SolarMassPanel.add(SolarMassTextField);
    SolarMassPanel.add(SolarMassUpdate);

    SolarMassButton1 = new JButton("<html>M<sub>\u2609</sub> = 1.98919\u00D710\u00B3\u00B3 (CoRoT/ESTA models)<html>");
    SolarMassButton1.addActionListener(SolarMassListener(1.98919e33));
    SolarMassButton1.setHorizontalAlignment(SwingConstants.LEFT);
    SolarMassButton2 = new JButton("<html>M<sub>\u2609</sub> = 1.989\u00D710\u00B3\u00B3 (Model S)<html>");
    SolarMassButton2.addActionListener(SolarMassListener(1.989e33));
    SolarMassButton2.setHorizontalAlignment(SwingConstants.LEFT);
    SolarMassButton3 = new JButton("<html>M<sub>\u2609</sub> = 1.98855\u00D710\u00B3\u00B3 (=GM<sub>\u2609</sub>/G, where GM<sub>\u2609</sub> comes from Cox 2000, and G from 2010 CODATA)<html>");
    SolarMassButton3.addActionListener(SolarMassListener(1.98855e33));
    SolarMassButton3.setHorizontalAlignment(SwingConstants.LEFT);
  }

  ////////////////////////////////////////////////////////////////////////////////////
  // This generates action listeners for the SolarMass buttons.
  ////////////////////////////////////////////////////////////////////////////////////
  private ActionListener SolarMassListener(final double value) {
    return new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        dataCenter.SolarMass = value;
        SolarMassCurrentValue.setText(String.format("Current value: %.15e (g)",dataCenter.SolarMass));
        SolarMassTextField.setText(String.format("%.15e",dataCenter.SolarMass));
      }
    };
  }

  ////////////////////////////////////////////////////////////////////////////////////
  // This initialises all the buttons used for setting the graviational constant
  // with different values found in the literature.
  ////////////////////////////////////////////////////////////////////////////////////
  private void initSolarRadiusButtons() {
    SolarRadiusLabel = new JLabel("<html><h2><font color='blue'>Solar radius, R<sub>\u2609</sub></font></h2></html>");
    SolarRadiusCurrentValue = new JLabel(String.format("Current value: %.15e (cm)",dataCenter.SolarRadius));

    SolarRadiusTextField = new JTextField(String.format("%.15e",dataCenter.SolarRadius));
    SolarRadiusTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        Double SolarRadius = Utilities.readDouble(SolarRadiusTextField.getText(),
                     "solar radius",6e10,8e10,SolarRadiusTextField);
        if (SolarRadius == null) {
          SolarRadiusTextField.setText(String.format("%.15e ",dataCenter.SolarRadius));
        } else {
          SolarRadiusTextField.setText(String.format("%.15e ",SolarRadius));
        }
      }
    });

    SolarRadiusUpdate = new JButton("Update value");
    SolarRadiusUpdate.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        Double SolarRadius = Utilities.readDouble(SolarRadiusTextField.getText(),
                     "solar radius",6e10,8e10,SolarRadiusUpdate);
        if (SolarRadius != null) {
          dataCenter.SolarRadius = SolarRadius;
          SolarRadiusCurrentValue.setText(String.format("Current value: %.15e (cm)",dataCenter.SolarRadius));
         }
        SolarRadiusTextField.setText(String.format("%.15e ",dataCenter.SolarRadius));
      }
    });

    SolarRadiusPanel = new JPanel(new GridLayout(1,2));
    SolarRadiusPanel.add(SolarRadiusTextField);
    SolarRadiusPanel.add(SolarRadiusUpdate);

    SolarRadiusButton1 = new JButton("<html>R<sub>\u2609</sub> = 6.9599\u00D710\u00B9\u2070 (CoRoT/ESTA models, Model S)<html>");
    SolarRadiusButton1.addActionListener(SolarRadiusListener(6.9599e10));
    SolarRadiusButton1.setHorizontalAlignment(SwingConstants.LEFT);
    SolarRadiusButton2 = new JButton("<html>R<sub>\u2609</sub> = 6.95613\u00D710\u00B9\u2070 (Haberreiter et al. 2008)<html>");
    SolarRadiusButton2.addActionListener(SolarRadiusListener(6.95613e10));
    SolarRadiusButton2.setHorizontalAlignment(SwingConstants.LEFT);
  }

  ////////////////////////////////////////////////////////////////////////////////////
  // This generates action listeners for the SolarRadius buttons.
  ////////////////////////////////////////////////////////////////////////////////////
  private ActionListener SolarRadiusListener(final double value) {
    return new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        dataCenter.SolarRadius = value;
        SolarRadiusCurrentValue.setText(String.format("Current value: %.15e (cm)",dataCenter.SolarRadius));
        SolarRadiusTextField.setText(String.format("%.15e",dataCenter.SolarRadius));
      }
    };
  }

  ////////////////////////////////////////////////////////////////////////////////////
  // This initialises SaveConstantsButton
  ////////////////////////////////////////////////////////////////////////////////////
  private void initSaveConstantsButton() {
    SaveConstantsButton = new JButton("Save constants");
    SaveConstantsButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent ev) {
        File f = new File(System.getProperty("user.home"),".InversionKit");
        try {
          PrintWriter out = new PrintWriter(new FileWriter(f));
          out.printf("%.15e # gravitational constant (in cm^3.g^-1.s^-2)%n",dataCenter.G);
          out.printf("%.15e # solar mass (in g)%n",dataCenter.SolarMass);
          out.printf("%.15e # solar radius (in cm)%n",dataCenter.SolarRadius);
          out.close();
          JOptionPane.showMessageDialog(
            SaveConstantsButton,
            "The constants were save in: "+f.getAbsolutePath(),
            "Constants saved",
            JOptionPane.INFORMATION_MESSAGE
          );
        } catch (Exception e) {
          JOptionPane.showMessageDialog(
            SaveConstantsButton,
            "Unable to write to: "+f.getAbsolutePath()+"\n"+
            "The constants were not saved.",
            "Error",
            JOptionPane.ERROR_MESSAGE
          );
        }
      }
    });
  }
}
