/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008, 2009, 2011, 2014
*
****************************************************************************/
package inversion_applet;

import inversion_applet.DataCenter;
import inversion_applet.FrequenciesDataModel;
import inversion_applet.Utilities;
import graph.IPlot;
import graph.Plot;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JCheckBox;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.SwingConstants;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.File;
import java.util.ArrayList;

public class FrequenciesPage extends JPanel {
  private JTable  frequenciesTable;
  private JScrollPane tableScroller;
  private JPanel buttonBox = new JPanel();
  private JButton findMatchButton;
  private JButton clearMatchButton;
  private JButton addRowButton;
  private JButton clearButton;
  private JButton sortButton;
  private JButton loadDataButton;
  private JButton appendDataButton;
  private JButton writeDataButton;
  private JButton writeFileButton;
  private JButton removeRowsButton;
  private JButton CalculateFreq;
  private JButton CalculateRS;
  private JButton plotTwoParamsButton;
  private JButton plotSplittingsConsistencyButton;
  private JButton plotSplittingsConsistencyR2Button;
  private JButton generateFrequenciesButton;
  private JButton replaceErrorFreqButton;
  private JButton replaceErrorRSButton;
  private JTextField SurfRotaTextField;
  private JTextField rminTextField;
  private JTextField gradientTextField;
  private JTextField gradientR2TextField;
  private double surfRotaValue;
  private double rminValue;
  private double gradientValue;
  private double gradientR2Value;
  private FrequenciesDataModel frequenciesData;
  private DataCenter dataCenter;

  // list of options in the two parameter plots
  private final String[] parameters = new String[]{"\u03BD(obs)","\u03BD(theo)",
                "E\u2099\u2097","RS","RS/\u222BK\u03A9","RS/\u222B(K\u03A9/r\u00B2)",
                "r\u2099\u2097","r'\u2099\u2097"};
  private final String[] parametersIDL = new String[]{"!7m!D!17obs!N","!7m!D!17theo!N",
                "E!Dn12l!17!N","RS","RS/!9i!17K!7!DX!N!17","RS/!9i!17(K!7!DX!N!17/r!U2!N)",
                "r!Dn!12l!17!N", "r''!Dn!12l!17!N"};

  ///////////////////////////////////////////////////////////
  // A constructor
  ///////////////////////////////////////////////////////////
  public FrequenciesPage(DataCenter dataCenter) {
    this.dataCenter = dataCenter;
  }

  ///////////////////////////////////////////////////////////
  // This creates the table with the frequencies
  ///////////////////////////////////////////////////////////
  private void createTable() {
    // Create the table data model:
    frequenciesData = new FrequenciesDataModel(dataCenter,buttonBox);
    dataCenter.frequenciesData = frequenciesData;

    // Create the table
    frequenciesTable = new JTable(frequenciesData);
    dataCenter.frequenciesTable = frequenciesTable;
    
    // Change the cell renderers:
    Utilities.IntegerRenderer ir1 = new Utilities.IntegerRenderer(Color.WHITE,Color.decode("#B8CFE5"));
    Utilities.IntegerRenderer ir2 = new Utilities.IntegerRenderer(Color.YELLOW,Color.decode("#B8CFE5"));
    frequenciesTable.getColumn("n").setCellRenderer(ir2);
    frequenciesTable.getColumn("\u2113").setCellRenderer(ir1);
    Utilities.DoubleRenderer dr1 = new Utilities.DoubleRenderer(3,Color.WHITE,Color.decode("#B8CFE5"));
    Utilities.DoubleRenderer dr2 = new Utilities.DoubleRenderer(3,Color.YELLOW,Color.decode("#B8CFE5"));
    frequenciesTable.getColumn("<html>\u03BD<sub>obs</sub> (\u03BCHz)</html>").setCellRenderer(dr1);
    frequenciesTable.getColumn("<html>\u03BD<sub>theo</sub> (\u03BCHz)</html>").setCellRenderer(dr2);
    frequenciesTable.getColumn("<html>\u03C3\u03BD<sub>obs</sub> (\u03BCHz)</html>").setCellRenderer(dr1);
    frequenciesTable.getColumn("Rot. Split.").setCellRenderer(dr1);
    frequenciesTable.getColumn("\u03C3(R.S.)").setCellRenderer(dr1);
    Utilities.BooleanRenderer br = new Utilities.BooleanRenderer(Color.YELLOW,Color.decode("#B8CFE5"));
    frequenciesTable.getColumn("Dupl.").setCellRenderer(br);

    // Create a scroll pane for the table
    tableScroller = new JScrollPane(frequenciesTable);
    return;
  }

  ///////////////////////////////////////////////////////////
  /** This sets up the user graphical interface */
  ///////////////////////////////////////////////////////////
  public void init() {
    initMatchButtons();
    initAddRowButton();
    initRemoveRowsButton();
    initClearButton();
    initSortButton();
    initReadDataButtons();
    initWriteDataButton();
    initWriteFileButton();
    initProposeModifyDataButtons();
    initCalculateDataButtons();
    initTestButtons();
    initTextFields();
    createTable();

    buttonBox = new JPanel();
    buttonBox.setLayout(new GridBagLayout());
    GridBagConstraints c = new GridBagConstraints();
    c.fill = GridBagConstraints.HORIZONTAL;
    c.gridx = 0;
    c.gridy = 0;
    c.gridwidth = 1;
    buttonBox.add(findMatchButton,c);
    c.gridx = 1;
    buttonBox.add(clearMatchButton,c);
    c.gridy++;
    c.gridx = 0;
    c.gridwidth = 2;
    buttonBox.add(new JLabel(" "),c);
    c.gridy++;
    c.gridwidth = 1;
    buttonBox.add(addRowButton,c);
    c.gridx = 1;
    buttonBox.add(removeRowsButton,c);
    c.gridy++;
    c.gridx = 0;
    buttonBox.add(clearButton,c);
    c.gridx = 1;
    buttonBox.add(sortButton,c);
    c.gridy++;
    c.gridx = 0;
    c.gridwidth = 2;
    buttonBox.add(new JLabel(" "),c);
    c.gridy++;
    c.gridwidth = 1;
    buttonBox.add(loadDataButton,c);
    c.gridx = 1;
    buttonBox.add(appendDataButton,c);
    c.gridy++;
    c.gridx = 0;
    buttonBox.add(writeDataButton,c);
    c.gridx = 1;
    buttonBox.add(writeFileButton,c);
    c.gridy++;
    c.gridx = 0;
    c.gridwidth = 2;
    buttonBox.add(new JLabel(" "),c);
    c.gridy++;
    buttonBox.add(generateFrequenciesButton,c);
    c.gridy++;
    buttonBox.add(replaceErrorFreqButton,c);
    c.gridy++;
    buttonBox.add(replaceErrorRSButton,c);
    c.gridy++;
    buttonBox.add(new JLabel(" "),c);
    c.gridy++;
    buttonBox.add(CalculateFreq,c);
    c.gridy++;
    buttonBox.add(CalculateRS,c);
    c.gridy++;
    buttonBox.add(new JLabel(" "),c);
    c.gridy++;  // createTable();
    buttonBox.add(frequenciesData.numberFreq,c);
    c.gridy++;
    buttonBox.add(frequenciesData.numberRS,c);
    c.gridy++;
    buttonBox.add(new JLabel(" "),c);
    c.gridy++;
    buttonBox.add(plotTwoParamsButton,c);
    c.gridy++;
    buttonBox.add(plotSplittingsConsistencyButton,c);
    c.gridy++;
    buttonBox.add(plotSplittingsConsistencyR2Button,c);
    c.gridy++;
    buttonBox.add(new JLabel(" "),c);
    c.gridy++;
    c.gridwidth = 1;
    c.gridx = 0;
    buttonBox.add(new JLabel("<html>&nbsp;\u03A9<sub>surface</sub>&nbsp;</html>"),c);
    c.gridx = 1;
    buttonBox.add(SurfRotaTextField,c);
    c.gridy++;
    c.gridx = 0;
    buttonBox.add(new JLabel("<html>&nbsp;r<sub>min</sub>&nbsp;</html>"),c);
    c.gridx = 1;
    buttonBox.add(rminTextField,c);
    c.gridy++;
    c.gridx = 0;
    buttonBox.add(new JLabel("<html>&nbsp;[-d\u03A9/dx]<sub>max</sub>&nbsp;</html>"),c);
    c.gridx = 1;
    buttonBox.add(gradientTextField,c);
    c.gridy++;
    c.gridx = 0;
    buttonBox.add(new JLabel("<html>&nbsp;[d(x\u00B2\u03A9)/dx]<sub>max</sub>&nbsp;</html>"),c);
    c.gridx = 1;
    buttonBox.add(gradientR2TextField,c);

    setLayout(new BorderLayout());
    add("Center",tableScroller);
    add("East",buttonBox);
  }

  ///////////////////////////////////////////////////////////
  // this initialises the find and clear match buttons
  ///////////////////////////////////////////////////////////
  private void initMatchButtons() {
    findMatchButton = Utilities.createButton("Find match");
    findMatchButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        dataCenter.findMatch();
        dataCenter.findDuplicates();
      }
    });

    clearMatchButton = Utilities.createButton("Clear match");
    clearMatchButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        dataCenter.clearMatch();
        dataCenter.clearDuplicates();
      }
    });
  }

  ///////////////////////////////////////////////////////////
  // this initialises addRowButton
  ///////////////////////////////////////////////////////////
  private void initAddRowButton() {
    addRowButton = Utilities.createButton("Add row");
    addRowButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        frequenciesData.addBlankRow(); // add a row
        frequenciesTable.updateUI();   // update the table
      }
    });
  }
  
  ///////////////////////////////////////////////////////////
  // this initialises removeRowsButton
  ///////////////////////////////////////////////////////////
  private void initRemoveRowsButton() {
    removeRowsButton = Utilities.createButton("Delete row(s)");
    removeRowsButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        // If a cell is being edited, the editing must first
        // stop before any rows can be removed.
        frequenciesTable.removeEditor();

        // find out which rows are selected
        int[] sr = frequenciesTable.getSelectedRows();

        // this stops ongoing cell editing:
        frequenciesTable.removeEditor();

        // This clears the selection (so as to avoid having
        // non-existent rows selected):
        frequenciesTable.clearSelection();

        if (sr.length == dataCenter.modes.size()) {
          // if the entire table is selected, just clear
          // the data (this is more rapid and sets the
          // correct value for dataCenter.DFloaded):
            dataCenter.clearModes();
        } else {
          // Remove the selected rows (the loop is backwards
          // to avoid problems with modified indices):
          for(int i = sr.length-1; i >= 0; i--) 
            frequenciesData.removeRow(sr[i]);
          // this updates the table:
          dataCenter.findDuplicates();
          dataCenter.kernelTable.updateUI();
        }

      }
    });
  }

  ///////////////////////////////////////////////////////////
  // this initialises clearButton
  ///////////////////////////////////////////////////////////
  private void initClearButton() {
    clearButton = Utilities.createButton("Clear");
    clearButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        frequenciesTable.removeEditor();
        frequenciesTable.clearSelection();
        dataCenter.clearModes(); // clears the data and updates tables
      }
    });
  }

  ///////////////////////////////////////////////////////////
  // this initialises sortButton
  ///////////////////////////////////////////////////////////
  private void initSortButton() {
    sortButton = Utilities.createButton("Sort");
    sortButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        frequenciesData.sort();       // sort the data
        frequenciesTable.updateUI();  // update the table
      }
    });
  }

  ///////////////////////////////////////////////////////////
  // this initialises loadDataButton
  ///////////////////////////////////////////////////////////
  private void initReadDataButtons() {
    appendDataButton = Utilities.createButton("Append data");
    appendDataButton.setEnabled(!dataCenter.isApplet);
    loadDataButton  = Utilities.createButton("Load data");
    loadDataButton.setEnabled(!dataCenter.isApplet);
    ActionListener l = new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        JButton source = (JButton) e.getSource();

        // open in current directory:
        JFileChooser fc = new JFileChooser(dataCenter.currentDirectory);

        // show the filechooser
        int result = fc.showOpenDialog(source);  // put parent component

        // get filename
        if(result == JFileChooser.APPROVE_OPTION) {
          File aFile = fc.getSelectedFile();

          // check to see if file exists:
          if (aFile.exists()) {
            if (source == loadDataButton) dataCenter.clearModes();
            frequenciesData.loadFile(aFile.getPath()); // load a file:
            frequenciesTable.updateUI();               // this updates the table:
            dataCenter.currentDirectory = aFile.getParentFile();
          } else {
            JOptionPane.showMessageDialog(
              source,
              "File \""+aFile.getName()+"\" doesn't exist.",
              "File not found",
              JOptionPane.WARNING_MESSAGE
            );
          }
        }
      }
    };
    loadDataButton.addActionListener(l);
    appendDataButton.addActionListener(l);
  }
  
  ///////////////////////////////////////////////////////////
  // this initialises writeDataButton
  ///////////////////////////////////////////////////////////
  private void initWriteDataButton() {
    writeDataButton  = Utilities.createButton("Write data");
    writeDataButton.setEnabled(!dataCenter.isApplet);
    writeDataButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {

        // open in current directory:
        JFileChooser fc = new JFileChooser(dataCenter.currentDirectory);

        // show the filechooser
        int result = fc.showSaveDialog(writeDataButton);  // put parent component

        // get filename
        if(result == JFileChooser.APPROVE_OPTION) {
          File aFile = fc.getSelectedFile();

          // check to see if file exists:
          if (aFile.exists()) {
            result = JOptionPane.showConfirmDialog(writeDataButton,
                 "Are you sure you want to overwrite "+aFile.getName()+"?",
                 "File exists",JOptionPane.YES_NO_OPTION);
          } else {
            result = JOptionPane.YES_OPTION;
          }

          if(result == JOptionPane.YES_OPTION)
            frequenciesData.writeFile(aFile.getPath());

          if (aFile.exists())
            dataCenter.currentDirectory = aFile.getParentFile();  
        }
      }
    });
  }

  ///////////////////////////////////////////////////////////
  // This recalculates the frequencies or rotational
  // splittings based on target structural or rotation
  // profiles
  ///////////////////////////////////////////////////////////
  private void initCalculateDataButtons() {
    CalculateFreq = Utilities.createButton("<html>Find \u03BD<sub>obs</sub> from targets</html>");
    CalculateFreq.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        frequenciesData.calculateFreq(CalculateFreq);
        frequenciesTable.updateUI();
      }
    });

    CalculateRS = Utilities.createButton("Find R.S. from target");
    CalculateRS.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        frequenciesData.calculateRS(CalculateRS);
        frequenciesTable.updateUI();
      }
    });
  }

  ///////////////////////////////////////////////////////////
  // This finds the number of modes with a rotational splitting,
  // a relevant l value, and a theoretical calculated mode.
  ///////////////////////////////////////////////////////////
  private int findNmodes() {
    int nmodes = 0, i;
    for(i=0;i<dataCenter.modes.size();i++) {
      if (checkMode(i)) nmodes++;
    }
    return nmodes;
  }

  ///////////////////////////////////////////////////////////
  // This finds the number of modes with a rotational splitting,
  // a relevant l value, and a theoretical calculated mode.
  ///////////////////////////////////////////////////////////
  private boolean checkMode(int i) {
    if (dataCenter.modes.get(i).l == null)  return false;
    if (dataCenter.modes.get(i).RS == null) return false;
    if (dataCenter.modes.get(i).kernel == null)  return false;
    if (dataCenter.modes.get(i).l.intValue() < 1) return false;
    return true;
  }

  ///////////////////////////////////////////////////////////
  // this initialises writeFileButton
  ///////////////////////////////////////////////////////////
  private void initWriteFileButton() {
    writeFileButton  = Utilities.createButton("Write file");
    writeFileButton.setEnabled(!dataCenter.isApplet);
    writeFileButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {

        // open in current directory:
        JFileChooser fc = new JFileChooser(dataCenter.currentDirectory);

        // show the filechooser
        int result = fc.showSaveDialog(writeFileButton);  // put parent component

        // get filename
        if(result == JFileChooser.APPROVE_OPTION) {
          File aFile = fc.getSelectedFile();

          // check to see if file exists:
          if (aFile.exists()) {
            result = JOptionPane.showConfirmDialog(writeFileButton,
                 "Are you sure you want to overwrite "+aFile.getName()+"?",
                 "File exists",JOptionPane.YES_NO_OPTION);
          } else {
            result = JOptionPane.YES_OPTION;
          }

          if(result == JOptionPane.YES_OPTION)
            writeFile(aFile.getPath());

          if (aFile.exists())
            dataCenter.currentDirectory = aFile.getParentFile();  
        }
      }
    });
  }

  /////////////////////////////////////////////////////////
  // write a file with rotational splittings, r_nl
  // and r'_nl for each mode
  /////////////////////////////////////////////////////////
  protected void writeFile(String filename) {
    File f = new File(filename);
    PrintWriter out;
    try {
      out = new PrintWriter(new FileWriter(f));
    } catch(Exception e) {
      JOptionPane.showMessageDialog(
        writeFileButton,
        "Unable to write to file \""+filename+"\"",
        "Failure",
        JOptionPane.WARNING_MESSAGE
      );
      return;
    }
    int ngrid = dataCenter.grid.length;
    out.printf("#%2s %3s %22s %22s %22s %22s %22s %22s %22s %22s %22s%n",
               "n","l","nu_obs (muHz)","nu_theo (muHz)","nu_var-nu_theo (muHz)",
               "Inertia", "RS", "Normalised RS", "Normalised RS (r^2)",
               "r_nl", "r'_nl");
    for(int i = 0; i < dataCenter.modes.size(); i++) {
      if (dataCenter.modes.get(i).kernel == null) continue;
      out.printf("%3d ",dataCenter.modes.get(i).kernel.n);
      out.printf("%3d ",dataCenter.modes.get(i).l);
      out.printf("%22.15e ",dataCenter.modes.get(i).FreqObs);
      out.printf("%22.15e ",dataCenter.modes.get(i).kernel.FreqTheo);
      out.printf("%22.15e ",dataCenter.modes.get(i).kernel.FreqVar-dataCenter.modes.get(i).kernel.FreqTheo);
      out.printf("%22.15e ",dataCenter.modes.get(i).kernel.IK);
      if (dataCenter.modes.get(i).RS != null) {
        out.printf("%22.15e ",dataCenter.modes.get(i).RS);
        out.printf("%22.15e ",dataCenter.modes.get(i).RS/dataCenter.modes.get(i).kernel.intRotaKernel[ngrid-1]);
        out.printf("%22.15e ",dataCenter.modes.get(i).RS/dataCenter.modes.get(i).kernel.intRotaKernelR2[ngrid-1]);
      } else {
        out.printf("%22s %22s %22s ","null","null","null");
      }
      out.printf("%22.15e ",dataCenter.modes.get(i).kernel.rRota);
      out.printf("%22.15e ",dataCenter.modes.get(i).kernel.rRotaR2);
      out.println();
    }
    out.close();
  }

  /////////////////////////////////////////////////////////////////////
  // This initialises the SurfRota text field
  /////////////////////////////////////////////////////////////////////
  private void initTextFields() {
    int nColumns = 10;
    surfRotaValue = 0.0;
    SurfRotaTextField = Utilities.createTextField(Double.toString(surfRotaValue));
    SurfRotaTextField.setColumns(nColumns);
    SurfRotaTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        checkSurfRotaTextField();
      }
    });

    rminValue = 0.0;
    rminTextField = Utilities.createTextField(Double.toString(rminValue));
    rminTextField.setColumns(nColumns);
    rminTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        checkRminTextField();
      }
    });

    gradientValue = 0.0;
    gradientTextField = Utilities.createTextField(Double.toString(gradientValue));
    gradientTextField.setColumns(nColumns);
    gradientTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        checkGradientTextField();
      }
    });

    gradientR2Value = 0.0;
    gradientR2TextField = Utilities.createTextField(Double.toString(gradientR2Value));
    gradientR2TextField.setColumns(nColumns);
    gradientR2TextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        checkGradientR2TextField();
      }
    });
  }

  /////////////////////////////////////////////////////////////////////
  // This checks the entry in SurfRotaTextField and updates
  // surfRotaValue if need be.
  //
  // Returns true if SurfRotaTextField contains a proper value
  //         false if otherwise
  /////////////////////////////////////////////////////////////////////
  private boolean checkSurfRotaTextField() {
    Double surfRota = Utilities.readDouble(SurfRotaTextField.getText(),
                      "surface rotation rate",0.0,1e10, SurfRotaTextField);
    if (surfRota == null) {
      SurfRotaTextField.setText(Double.toString(surfRotaValue));
      return false; // no need to put a second error message
    }
    surfRotaValue = surfRota.doubleValue();
    return true;
  }

  /////////////////////////////////////////////////////////////////////
  // This checks the entry in rminTextField and updates
  // rminValue if need be.
  //
  // Returns true if rminTextField contains a proper value
  //         false if otherwise
  /////////////////////////////////////////////////////////////////////
  private boolean checkRminTextField() {
    Double rmin = null;
    if (dataCenter.grid != null) {
      rmin = Utilities.readDouble(rminTextField.getText(),
             "lower bound",0.0,dataCenter.grid[dataCenter.grid.length-1], rminTextField);
    } else {
      rmin = Utilities.readDouble(rminTextField.getText(),"lower bound",0.0,1.0, rminTextField);
    }
    if (rmin == null) {
      rminTextField.setText(Double.toString(rminValue));
      return false; // no need to put a second error message
    }
    rminValue = rmin.doubleValue();
    return true;
  }

  /////////////////////////////////////////////////////////////////////
  // This checks the entry in gradientTextField and updates
  // gradientValue if need be.
  //
  // Returns true if gradientTextField contains a proper value
  //         false if otherwise
  /////////////////////////////////////////////////////////////////////
  private boolean checkGradientTextField() {
    Double gradient = Utilities.readDouble(gradientTextField.getText(),
                  "lower bound on the gradient of \u03A9 in the center",0.0,1e10, gradientTextField);
    if (gradient == null) {
      gradientTextField.setText(Double.toString(gradientValue));
      return false; // no need to put a second error message
    }
    gradientValue = gradient.doubleValue();
    return true;
  }

  /////////////////////////////////////////////////////////////////////
  // This checks the entry in gradientR2TextField and updates
  // gradientR2Value if need be.
  //
  // Returns true if gradientR2TextField contains a proper value
  //         false if otherwise
  /////////////////////////////////////////////////////////////////////
  private boolean checkGradientR2TextField() {
    Double gradient = Utilities.readDouble(gradientR2TextField.getText(),
                  "upper bound on the gradient of (x\u00B2\u03A9) in the center",0.0,1e10, gradientR2TextField);
    if (gradient == null) {
      gradientR2TextField.setText(Double.toString(gradientR2Value));
      return false; // no need to put a second error message
    }
    gradientR2Value = gradient.doubleValue();
    return true;
  }

  ///////////////////////////////////////////////////////////
  // This sets up buttons which modify the error bars or
  // proposes a grid of frequencies.
  ///////////////////////////////////////////////////////////
  private void initProposeModifyDataButtons() {
    generateFrequenciesButton = Utilities.createButton("Generate frequencies");
    generateFrequenciesButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {

        JTextField lStartField = new JTextField("0");
        JTextField lStopField  = new JTextField("3");
        JTextField lIncrField  = new JTextField("1");
        JTextField freqStartField = new JTextField();
        JTextField freqStopField  = new JTextField();
        JTextField freqIncrField  = new JTextField();

        JLabel lLabel = new JLabel("\u2113 values:");
        JLabel lStartLabel = new JLabel("Initial value: ",SwingConstants.RIGHT);
        JLabel lStopLabel  = new JLabel("Final value: ",SwingConstants.RIGHT);
        JLabel lIncrLabel  = new JLabel("Increment: ",SwingConstants.RIGHT);
        JLabel freqLabel = new JLabel("Frequency values (in \u03BCHz):");
        JLabel freqStartLabel = new JLabel("Initial value: ",SwingConstants.RIGHT);
        JLabel freqStopLabel  = new JLabel("Final value: ",SwingConstants.RIGHT);
        JLabel freqIncrLabel  = new JLabel("Increment: ",SwingConstants.RIGHT);

        JPanel lStartPanel = new JPanel(new GridLayout(1,2));
        JPanel lStopPanel  = new JPanel(new GridLayout(1,2));
        JPanel lIncrPanel  = new JPanel(new GridLayout(1,2));
        JPanel freqStartPanel = new JPanel(new GridLayout(1,2));
        JPanel freqStopPanel  = new JPanel(new GridLayout(1,2));
        JPanel freqIncrPanel  = new JPanel(new GridLayout(1,2));

        lStartPanel.add(lStartLabel);
        lStartPanel.add(lStartField);
        lStopPanel.add(lStopLabel);
        lStopPanel.add(lStopField);
        lIncrPanel.add(lIncrLabel);
        lIncrPanel.add(lIncrField);

        freqStartPanel.add(freqStartLabel);
        freqStartPanel.add(freqStartField);
        freqStopPanel.add(freqStopLabel);
        freqStopPanel.add(freqStopField);
        freqIncrPanel.add(freqIncrLabel);
        freqIncrPanel.add(freqIncrField);

        JLabel ErrorFreqLabel = new JLabel("<html>\u03C3\u03BD<sub>obs</sub></html>",SwingConstants.RIGHT);
        JTextField ErrorFreqField = new JTextField();
        JPanel ErrorFreqPanel = new JPanel(new GridLayout(1,2));
        ErrorFreqPanel.add(ErrorFreqLabel);
        ErrorFreqPanel.add(ErrorFreqField);

        final Object[] message = new Object[13];
        message[0] = "<html><h2><font color='blue'>Generate frequencies</font></h2></html>";
        message[1] = " ";
        message[2] = freqLabel;
        message[3] = freqStartPanel;
        message[4] = freqStopPanel;
        message[5] = freqIncrPanel;
        message[6] = " ";
        message[7] = lLabel;
        message[8] = lStartPanel;
        message[9] = lStopPanel;
        message[10] = lIncrPanel;
        message[11] = " ";
        message[12] = ErrorFreqPanel;

        String[] options = new String[2];
        options[0] = "Generate frequencies";
        options[1] = "Cancel";

        int result = JOptionPane.showOptionDialog(
          generateFrequenciesButton,    // the parent that the dialog blocks
          message,                      // the dialog message array
          "Generate frequencies",       // the title of the dialog window
          JOptionPane.DEFAULT_OPTION,   // option type
          JOptionPane.QUESTION_MESSAGE, // message type
          null,                         // optional icon, use null to use the default icon
          options,                      // options string array, will be made into buttons
          options[0]                    // option that should be made into a default button
        );
        
        switch(result) {
          case 1:  // Cancel
            break;

          default: // generate frequencies:

            // Check to see if different entries are acceptable:
            Double freqStart = Utilities.readDouble(freqStartField.getText(),
                    "initial frequency",0.0,1e6,freqStartField);
            if (freqStart == null) break;
            Double freqStop = Utilities.readDouble(freqStopField.getText(),
                    "final frequency",0.0,1e6,freqStopField);
            if (freqStop == null) break;
            Double freqIncr = Utilities.readDouble(freqIncrField.getText(),
                    "frequency increment",-1e6,1e6,freqIncrField);
            if (freqIncr == null) break;

            Integer lStart = Utilities.readInt(lStartField.getText(),
                    "initial \u2113 value",0,1000000,lStartField);
            if (lStart == null) break;
            Integer lStop = Utilities.readInt(lStopField.getText(),
                    "final \u2113 value",0,1000000,lStopField);
            if (lStop == null) break;
            Integer lIncr = Utilities.readInt(lIncrField.getText(),
                    "\u2113 increment",-1000000,1000000,lIncrField);
            if (lIncr == null) break;

            Double ErrorFreq = Utilities.readDouble(ErrorFreqField.getText(),
                    "initial frequency",0.0,1e6,ErrorFreqField);
            if (ErrorFreq == null) break;

            // sanity checks
            if (((freqStop.doubleValue()-freqStart.doubleValue())*freqIncr.doubleValue() < 0.0)||
                 (freqIncr.doubleValue() == 0.0)) {
              JOptionPane.showMessageDialog(
                generateFrequenciesButton,
                "Unable to generate frequencies due to\n"+
                "an inconsistent frequency increment.",
                "Error",
                JOptionPane.ERROR_MESSAGE
              );
              break;
            }

            int dl = lStop.intValue() - lStart.intValue();
            if (((dl < 0)&&(lIncr.intValue() > 0))||
                ((dl > 0)&&(lIncr.intValue() < 0))||
                 (lIncr.intValue() == 0)) {
              JOptionPane.showMessageDialog(
                generateFrequenciesButton,
                "Unable to generate frequencies due to\n"+
                "an inconsistent \u2113 increment.",
                "Error",
                JOptionPane.ERROR_MESSAGE
              );
              break;
            }

            frequenciesData.generateFrequencies(
              lStart.intValue(), lStop.intValue(), lIncr.intValue(),
              freqStart.doubleValue(), freqStop.doubleValue(), freqIncr.doubleValue(),
              ErrorFreq
            );
            frequenciesTable.updateUI();
            break;
        }
      }
    });

    replaceErrorFreqButton = Utilities.createButton("<html>Modify \u03C3\u03BD<sub>obs</sub></html>");
    replaceErrorFreqButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        String strErrorFreq = JOptionPane.showInputDialog(
          replaceErrorFreqButton,
          "<html>Please enter a new value for \u03C3\u03BD<sub>obs</sub> (in \u03BCHz):</html>",
          "Modify \u03C3\u03BD(obs)",
          JOptionPane.QUESTION_MESSAGE
        );
        if (strErrorFreq == null) return;
        Double ErrorFreq = Utilities.readDouble(strErrorFreq,"\u03C3\u03BD(obs)",0.0,1e10,replaceErrorFreqButton);
        if (ErrorFreq == null) return;
        frequenciesData.replaceErrorFreq(ErrorFreq);
        frequenciesTable.updateUI();
      }
    });

    replaceErrorRSButton = Utilities.createButton("Modify \u03C3(R.S.)");
    replaceErrorRSButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        String strErrorRS = JOptionPane.showInputDialog(
          replaceErrorRSButton,
          "<html>Please enter a new value for \u03C3(R.S.) (in \u03BCHz):</html>",
          "Modify \u03C3(R.S.)",
          JOptionPane.QUESTION_MESSAGE
        );
        if (strErrorRS == null) return;
        Double ErrorRS = Utilities.readDouble(strErrorRS,"\u03C3(R.S.)",0.0,1e10,replaceErrorRSButton);
        if (ErrorRS == null) return;
        frequenciesData.replaceErrorRS(ErrorRS);
        frequenciesTable.updateUI();
      }
    });
  }

  ///////////////////////////////////////////////////////////
  // This produces an array of values based on the choice of
  // a parameter
  ///////////////////////////////////////////////////////////
  private double[] parameterToValues(String parameter, int[] selection) {
    double[] values = new double[selection.length];
    int i,ngrid=dataCenter.grid.length;
    if (parameter.equals("\u03BD(obs)")) {
      for(i=0;i<selection.length;i++) {
        values[i] = Utilities.DoubleTodouble(dataCenter.modes.get(selection[i]).FreqObs);
      }
    } else if (parameter.equals("\u03BD(theo)")) {
      for(i=0;i<selection.length;i++) {
        if (dataCenter.modes.get(selection[i]).kernel == null) {
          values[i] = Double.NaN;
        } else {
          values[i] = dataCenter.modes.get(selection[i]).kernel.FreqTheo;
        }
      }
    } else if (parameter.equals("E\u2099\u2097")) {
      for(i=0;i<selection.length;i++) {
        if (dataCenter.modes.get(selection[i]).kernel == null) {
          values[i] = Double.NaN;
        } else {
          values[i] = dataCenter.modes.get(selection[i]).kernel.Enl;
        }
      }
    } else if (parameter.equals("RS")) {
      for(i=0;i<selection.length;i++) {
        values[i] = Utilities.DoubleTodouble(dataCenter.modes.get(selection[i]).RS);
      }
    } else if (parameter.equals("RS/\u222BK\u03A9")) {
      for(i=0;i<selection.length;i++) {
        if (dataCenter.modes.get(selection[i]).kernel == null) {
          values[i] = Double.NaN;
        } else {
          values[i] = Utilities.DoubleTodouble(dataCenter.modes.get(selection[i]).RS)
                    / dataCenter.modes.get(selection[i]).kernel.intRotaKernel[ngrid-1];
        }
      }
    } else if (parameter.equals("RS/\u222B(K\u03A9/r\u00B2)")) {
      for(i=0;i<selection.length;i++) {
        if (dataCenter.modes.get(selection[i]).kernel == null) {
          values[i] = Double.NaN;
        } else {
          values[i] = Utilities.DoubleTodouble(dataCenter.modes.get(selection[i]).RS)
                    / dataCenter.modes.get(selection[i]).kernel.intRotaKernelR2[ngrid-1];
        }
      }
    } else if (parameter.equals("r\u2099\u2097")) {
      for(i=0;i<selection.length;i++) {
        if (dataCenter.modes.get(selection[i]).kernel == null) {
          values[i] = Double.NaN;
        } else {
          values[i] = dataCenter.modes.get(selection[i]).kernel.rRota;
        }
      }
    } else if (parameter.equals("r'\u2099\u2097")) {
      for(i=0;i<selection.length;i++) {
        if (dataCenter.modes.get(selection[i]).kernel == null) {
          values[i] = Double.NaN;
        } else {
          values[i] = dataCenter.modes.get(selection[i]).kernel.rRotaR2;
        }
      }
    } else {
      return null;
    }
    return values;
  }

  ///////////////////////////////////////////////////////////
  // This produces an array of error bars based on the choice of
  // a parameter
  ///////////////////////////////////////////////////////////
  private double[] parameterToErrors(String parameter, int[] selection) {
    double[] values = new double[selection.length];
    int i,ngrid=dataCenter.grid.length;
    if (parameter.equals("\u03BD(obs)")) {
      for(i=0;i<selection.length;i++) {
        values[i] = Utilities.DoubleTodouble(dataCenter.modes.get(selection[i]).ErrorFreq);
      }
    } else if (parameter.equals("\u03BD(theo)")) {
      return null;
    } else if (parameter.equals("E\u2099\u2097")) {
      return null;
    } else if (parameter.equals("RS")) {
      for(i=0;i<selection.length;i++) {
        values[i] = Utilities.DoubleTodouble(dataCenter.modes.get(selection[i]).ErrorRS);
      }
    } else if (parameter.equals("RS/\u222BK\u03A9")) {
      for(i=0;i<selection.length;i++) {
        if (dataCenter.modes.get(selection[i]).kernel == null) {
          values[i] = Double.NaN;
        } else {
          values[i] = Utilities.DoubleTodouble(dataCenter.modes.get(selection[i]).ErrorRS)
                    / dataCenter.modes.get(selection[i]).kernel.intRotaKernel[ngrid-1];
        }
      }
    } else if (parameter.equals("RS/\u222B(K\u03A9/r\u00B2)")) {
      for(i=0;i<selection.length;i++) {
        if (dataCenter.modes.get(selection[i]).kernel == null) {
          values[i] = Double.NaN;
        } else {
          values[i] = Utilities.DoubleTodouble(dataCenter.modes.get(selection[i]).ErrorRS)
                    / dataCenter.modes.get(selection[i]).kernel.intRotaKernelR2[ngrid-1];
        }
      }
    } else if (parameter.equals("r\u2099\u2097")) {
      return null;
    } else if (parameter.equals("r'\u2099\u2097")) {
      return null;
    } else {
      return null;
    }
    return values;
  }

  ///////////////////////////////////////////////////////////
  // this initialises various buttons related to testing
  // the rotational kernels or plotting various mode
  // quantities.
  ///////////////////////////////////////////////////////////
  private void initTestButtons() {
    plotTwoParamsButton = Utilities.createButton("Plot 2 parameters");
    plotTwoParamsButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (!dataCenter.modelLoaded) {
          JOptionPane.showMessageDialog(
            plotTwoParamsButton,
            "No model loaded, please load the model first.",
            "No model",
            JOptionPane.ERROR_MESSAGE
          );
          return;
        }

        if (!checkSurfRotaTextField()) return;
        if (!checkRminTextField()) return;
        if (!checkGradientTextField()) return;
        if (!checkGradientR2TextField()) return;

        final JComboBox xParam = Utilities.createComboBox(parameters);
        final JComboBox yParam = Utilities.createComboBox(parameters);
        final JCheckBox withErrorsCheckBox = new JCheckBox("with error bars",false);
        final JCheckBox inconsistencyCheckBox = new JCheckBox("RS inconsis.",false);
        final JCheckBox inconsistencyR2CheckBox = new JCheckBox("RS inconsis. R\u00B2",false);
        final int[] selection = frequenciesTable.getSelectedRows();
        xParam.setSelectedItem("RS/\u222BK\u03A9");
        yParam.setSelectedItem("r\u2099\u2097");
        String xSelection = (String)xParam.getSelectedItem();
        String ySelection = (String)yParam.getSelectedItem();
        String xSelectionIDL = parametersIDL[xParam.getSelectedIndex()];
        String ySelectionIDL = parametersIDL[yParam.getSelectedIndex()];
        double[] xArray = parameterToValues(xSelection,selection);
        double[] yArray = parameterToValues(ySelection,selection);
        final IPlot aPlot = new IPlot(ySelection+" vs. "+xSelection,xSelection,ySelection);
        aPlot.setIDLTitles(ySelectionIDL+" vs. "+xSelectionIDL,xSelectionIDL,ySelectionIDL);
        aPlot.addDataSet(0,xArray,yArray,Plot.noLine,Plot.plus,Color.blue);
        aPlot.findRange();

        ActionListener l = new ActionListener() {
          public void actionPerformed(ActionEvent ev) {
            int i,j,ii,jj;
            int ngrid = dataCenter.grid.length;
            double R = dataCenter.grid[ngrid-1];
            double surfRota = Utilities.parseDouble(SurfRotaTextField.getText());
            double[][][] bounds;
            double[] xx = new double[2];
            double[] yy = new double[2];
            Double ErrorRSi, ErrorRSj;
            String xSelection = (String)xParam.getSelectedItem();
            String ySelection = (String)yParam.getSelectedItem();
            String xSelectionIDL = parametersIDL[xParam.getSelectedIndex()];
            String ySelectionIDL = parametersIDL[yParam.getSelectedIndex()];
            double[] xArray = parameterToValues(xSelection,selection);
            double[] yArray = parameterToValues(ySelection,selection);
            aPlot.clearDataSets();
            aPlot.setTitles(ySelection+" vs. "+xSelection,xSelection,ySelection);
            aPlot.setIDLTitles(ySelectionIDL+" vs. "+xSelectionIDL,xSelectionIDL,ySelectionIDL);

            if (inconsistencyCheckBox.isSelected()) {
              bounds = findBounds(rminValue);
              for(ii=0;ii<selection.length;ii++) {
                i = selection[ii];
                ErrorRSi = dataCenter.modes.get(i).ErrorRS;
                if (dataCenter.modes.get(i).RS == null) continue;
                if (ErrorRSi == null) continue;
                if (dataCenter.modes.get(i).kernel == null) continue;
                for(jj=ii+1;jj<selection.length;jj++) {
                  j = selection[jj];
                  ErrorRSj = dataCenter.modes.get(j).ErrorRS;
                  if (dataCenter.modes.get(j).RS == null) continue;
                  if (ErrorRSj == null) continue;
                  if (dataCenter.modes.get(j).kernel == null) continue;
                  if (Double.isNaN(bounds[0][i][j])) continue;
                  if (Double.isNaN(bounds[1][i][j])) continue;
                  if ((bounds[0][i][j]*(dataCenter.modes.get(j).RS-ErrorRSj)/dataCenter.modes.get(j).kernel.intRotaKernel[ngrid-1]-gradientValue*bounds[2][i][j]
                                   <=  (dataCenter.modes.get(i).RS+ErrorRSi)/dataCenter.modes.get(i).kernel.intRotaKernel[ngrid-1]) &&
                                      ((dataCenter.modes.get(i).RS-ErrorRSi)/dataCenter.modes.get(i).kernel.intRotaKernel[ngrid-1] 
                   <=  bounds[1][i][j]*(dataCenter.modes.get(j).RS+ErrorRSj)/dataCenter.modes.get(j).kernel.intRotaKernel[ngrid-1]+gradientValue*bounds[3][i][j])) continue;
                  xx[0] = xArray[ii];
                  xx[1] = xArray[jj];
                  yy[0] = yArray[ii];
                  yy[1] = yArray[jj];
                  aPlot.addDataSet(xx,yy,Plot.dotted,Plot.noSymbol,Color.green);
                }
              }
            }

            if (inconsistencyR2CheckBox.isSelected()) {
              bounds = findBoundsR2(rminValue);
              for(ii=0;ii<selection.length;ii++) {
                i = selection[ii];
                ErrorRSi = dataCenter.modes.get(i).ErrorRS;
                if (dataCenter.modes.get(i).RS == null) continue;
                if (ErrorRSi == null) continue;
                if (dataCenter.modes.get(i).kernel == null) continue;
                for(jj=ii+1;jj<selection.length;jj++) {
                  j = selection[jj];
                  ErrorRSj = dataCenter.modes.get(j).ErrorRS;
                  if (dataCenter.modes.get(j).RS == null) continue;
                  if (ErrorRSj == null) continue;
                  if (dataCenter.modes.get(j).kernel == null) continue;
                  if (Double.isNaN(bounds[0][i][j])) continue;
                  if (Double.isNaN(bounds[1][i][j])) continue;
                  if (((bounds[1][i][j]*((dataCenter.modes.get(j).RS-ErrorRSj)/dataCenter.modes.get(j).kernel.intRotaKernelR2[ngrid-1]-surfRota*R*R)-gradientR2Value*bounds[3][i][j])
                                    <=  ((dataCenter.modes.get(i).RS+ErrorRSi)/dataCenter.modes.get(i).kernel.intRotaKernelR2[ngrid-1]-surfRota*R*R)) &&
                                       (((dataCenter.modes.get(i).RS-ErrorRSi)/dataCenter.modes.get(i).kernel.intRotaKernelR2[ngrid-1]-surfRota*R*R) 
                   <=  (bounds[0][i][j]*((dataCenter.modes.get(j).RS+ErrorRSj)/dataCenter.modes.get(j).kernel.intRotaKernelR2[ngrid-1]-surfRota*R*R)-gradientR2Value*bounds[2][i][j]))) continue;
                  xx[0] = xArray[ii];
                  xx[1] = xArray[jj];
                  yy[0] = yArray[ii];
                  yy[1] = yArray[jj];
                  aPlot.addDataSet(xx,yy,Plot.dashed,Plot.noSymbol,Color.red);
                }
              }
            }

            aPlot.addDataSet(xArray,yArray,Plot.noLine,Plot.plus,Color.blue);
            int n = aPlot.getNDataSets();
            if (withErrorsCheckBox.isSelected()) {
              double[] xErrors = parameterToErrors(xSelection,selection);
              double[] yErrors = parameterToErrors(ySelection,selection);
              if (xErrors != null) {
                aPlot.replaceLeftError(n-1, xErrors);
                aPlot.replaceRightError(n-1, xErrors);
                aPlot.setHErrorValue(n-1,true);
              } else {
                aPlot.setHErrorValue(n-1,false);
              }
              if (yErrors != null) {
                aPlot.replaceLowerError(n-1, yErrors);
                aPlot.replaceUpperError(n-1, yErrors);
                aPlot.setVErrorValue(n-1,true);
              } else {
                aPlot.setVErrorValue(n-1,false);
              }
            } else {
              aPlot.setHErrorValue(n-1,false);
              aPlot.setVErrorValue(n-1,false);
            }
            aPlot.findRange();
            aPlot.repaint();
          }
        };
        xParam.addActionListener(l);
        yParam.addActionListener(l);
        withErrorsCheckBox.addActionListener(l);
        inconsistencyCheckBox.addActionListener(l);
        inconsistencyR2CheckBox.addActionListener(l);

        JFrame f = new JFrame("Two parameters");
        f.setLayout(new BorderLayout());
        f.add("Center",aPlot);
        JPanel buttonBox = new JPanel();
        buttonBox.setLayout(new GridBagLayout());
        GridBagConstraints c = new GridBagConstraints();
        c.fill = GridBagConstraints.HORIZONTAL;
        c.gridx = 0;
        c.gridy = 0;
        c.gridwidth = 1;
        buttonBox.add(new JLabel("<html><font color='blue'><b>&nbsp;&nbsp;&nbsp;x-axis&nbsp;&nbsp;&nbsp;</b></font></html>"),c);
        c.gridy++;
        buttonBox.add(xParam,c);
        c.gridy++;
        buttonBox.add(new JLabel(" "),c);
        c.gridy++;
        buttonBox.add(new JLabel("<html><font color='blue'><b>&nbsp;&nbsp;&nbsp;y-axis&nbsp;&nbsp;&nbsp;</b></font></html>"),c);
        c.gridy++;
        buttonBox.add(yParam,c);
        c.gridy++;
        buttonBox.add(new JLabel(" "),c);
        c.gridy++;
        buttonBox.add(withErrorsCheckBox,c);
        c.gridy++;
        buttonBox.add(inconsistencyCheckBox,c);
        c.gridy++;
        buttonBox.add(inconsistencyR2CheckBox,c);
        c.gridy++;
        buttonBox.add(new JLabel("<html>\u03A9<sub>surface</sub>="+surfRotaValue+"</html>"),c);
        c.gridy++;
        buttonBox.add(new JLabel("<html>r<sub>min</sub>="+rminValue+"</html>"),c);
        c.gridy++;
        buttonBox.add(new JLabel("<html>[-d\u03A9/dx]<sub>max</sub>="+gradientValue+"</html>"),c);
        c.gridy++;
        buttonBox.add(new JLabel("<html>[d(x\u00B2\u03A9)/dx]<sub>max</sub>="+gradientR2Value+"</html>"),c);
        f.add("East",buttonBox);
        f.setSize(600,600);
        f.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
        f.setVisible(true);    
      }
    });
   
    plotSplittingsConsistencyButton = Utilities.createButton("Plot splittings test");
    plotSplittingsConsistencyButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (!checkRminTextField()) return;
        if (!checkGradientTextField()) return;

        plotSplittingsConsistency();
      }
    });

    plotSplittingsConsistencyR2Button = Utilities.createButton("Plot splittings R\u00B2 test");
    plotSplittingsConsistencyR2Button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (!checkSurfRotaTextField()) return;
        if (!checkRminTextField()) return;
        if (!checkGradientR2TextField()) return;

        plotSplittingsConsistencyR2(Utilities.parseDouble(SurfRotaTextField.getText()));
      }
    });
  }

  ///////////////////////////////////////////////////////////
  // This produces a series of plots to check the consistency
  // of the rotational splittings with the rotation kernels,
  // assuming a rotation profile with a negative gradient.
  ///////////////////////////////////////////////////////////
  private void plotSplittingsConsistency() {
    if (!dataCenter.modelLoaded) {
      JOptionPane.showMessageDialog(
        plotSplittingsConsistencyButton,
        "No model loaded, please load the model first.",
        "No model",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }

    int i, j, k, size;
    int ngrid = dataCenter.grid.length;
    double cnst, lfreq, ufreq, dfreq;
    final double[][][] bounds = findBounds(rminValue);
    ArrayList<Double> lowerValues = new ArrayList<Double>();
    ArrayList<Double> upperValues = new ArrayList<Double>();
    ArrayList<Double> freq        = new ArrayList<Double>();
    double[] x, y, error;
    Double ErrorRSi, ErrorRSj;
    JTabbedPane tabbedPane = new JTabbedPane();
    String title, IDLtitle;
    IPlot onePlot;

    for(i=0; i<dataCenter.modes.size(); i++) {
      ErrorRSi = dataCenter.modes.get(i).ErrorRS;
      if (dataCenter.modes.get(i).l.intValue() < 1) continue;
      if (dataCenter.modes.get(i).RS == null) continue;
      if (ErrorRSi == null) continue;
      if (dataCenter.modes.get(i).kernel == null) continue;
      lowerValues.clear();
      upperValues.clear();
      freq.clear();
      lfreq = Double.POSITIVE_INFINITY;
      ufreq = Double.NEGATIVE_INFINITY;
      for(j=0; j<dataCenter.modes.size(); j++) {
        ErrorRSj = dataCenter.modes.get(j).ErrorRS;
        if (Double.isNaN(bounds[0][i][j])) continue; 
        if (Double.isNaN(bounds[1][i][j])) continue; 
        if (dataCenter.modes.get(j).RS == null) continue;
        if (ErrorRSj == null) continue;
        if (dataCenter.modes.get(j).kernel == null) continue;
        if (i==j) continue;
        
        lowerValues.add((dataCenter.modes.get(j).RS-ErrorRSj)
                *bounds[0][i][j]/dataCenter.modes.get(j).kernel.intRotaKernel[ngrid-1]
                -gradientValue*bounds[2][i][j]);
        upperValues.add((dataCenter.modes.get(j).RS+ErrorRSj)
                *bounds[1][i][j]/dataCenter.modes.get(j).kernel.intRotaKernel[ngrid-1]
                +gradientValue*bounds[3][i][j]);
        freq.add(dataCenter.modes.get(j).FreqObs);
        if (lfreq > dataCenter.modes.get(j).FreqObs.doubleValue())
            lfreq = dataCenter.modes.get(j).FreqObs.doubleValue();
        if (ufreq < dataCenter.modes.get(j).FreqObs.doubleValue())
            ufreq = dataCenter.modes.get(j).FreqObs.doubleValue();
      }
      if (lfreq > dataCenter.modes.get(i).FreqObs.doubleValue())
          lfreq = dataCenter.modes.get(i).FreqObs.doubleValue();
      if (ufreq < dataCenter.modes.get(i).FreqObs.doubleValue())
          ufreq = dataCenter.modes.get(i).FreqObs.doubleValue();
      dfreq = (ufreq-lfreq)*0.1;
      lfreq-=dfreq;
      ufreq+=dfreq;

      // produce plot:
      title = "(n,\u2113)=("+dataCenter.modes.get(i).kernel.n
                            +","+dataCenter.modes.get(i).l.intValue()+")";
      IDLtitle = "!17(n,!12l!17)=("+dataCenter.modes.get(i).kernel.n
                               +","+dataCenter.modes.get(i).l.intValue()+")";
      onePlot = new IPlot(title,"Frequency (in \u03BCHz)","Normalised splitting");
      onePlot.setIDLTitles(IDLtitle,"!17Frequency (in !7l!17Hz)","Normalised splitting");
      size = lowerValues.size(); 

      // the original splittings
      x = new double[]{lfreq,ufreq};
      cnst = dataCenter.modes.get(i).RS/dataCenter.modes.get(i).kernel.intRotaKernel[ngrid-1];
      y = new double[]{cnst,cnst};
      onePlot.addDataSet(x, y, Plot.solid, Plot.noSymbol, Color.blue);
      cnst = (dataCenter.modes.get(i).RS+ErrorRSi)
           /  dataCenter.modes.get(i).kernel.intRotaKernel[ngrid-1];
      y = new double[]{cnst,cnst};
      onePlot.addDataSet(x, y, Plot.dotted, Plot.noSymbol, Color.blue);
      cnst = (dataCenter.modes.get(i).RS-ErrorRSi)
           /  dataCenter.modes.get(i).kernel.intRotaKernel[ngrid-1];
      y = new double[]{cnst,cnst};
      onePlot.addDataSet(x, y, Plot.dotted, Plot.noSymbol, Color.blue);
      x = new double[]{dataCenter.modes.get(i).FreqObs};
      y = new double[]{dataCenter.modes.get(i).RS/dataCenter.modes.get(i).kernel.intRotaKernel[ngrid-1]};
      error = new double[]{ErrorRSi/dataCenter.modes.get(i).kernel.intRotaKernel[ngrid-1]};
      onePlot.addDataSet(x, y, Plot.noLine, Plot.noSymbol, Color.red);
      onePlot.replaceLowerError(3,error); 
      onePlot.replaceUpperError(3,error); 
      onePlot.setVErrorValue(3,true);

      // the constraints from the other splittings
      for(j=0;j<lowerValues.size();j++) {
        if ((lowerValues.get(j)==0.0)&&(Double.isInfinite(upperValues.get(j)))) continue;
        x = new double[]{freq.get(j),freq.get(j)};
        y = new double[]{lowerValues.get(j),upperValues.get(j)};
        onePlot.addDataSet(x, y, Plot.solid, Plot.times, Color.blue);
      }
      onePlot.findRange();

      title = "("+dataCenter.modes.get(i).kernel.n
             +","+dataCenter.modes.get(i).l.intValue()+")";
      tabbedPane.add(title,onePlot);
    }

    final JButton printBounds = new JButton("Print bounds (in LaTeX format)");
    printBounds.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        // open in current directory:
        JFileChooser fc = new JFileChooser(dataCenter.currentDirectory);

        // show the filechooser
        int result = fc.showSaveDialog(printBounds);  // put parent component

        // get filename
        if(result == JFileChooser.APPROVE_OPTION) {
          File aFile = fc.getSelectedFile();

          // check to see if file exists:
          if (aFile.exists()) {
            result = JOptionPane.showConfirmDialog(printBounds,
                 "Are you sure you want to overwrite "+aFile.getName()+"?",
                 "File exists",JOptionPane.YES_NO_OPTION);
          } else {
            result = JOptionPane.YES_OPTION;
          }

          if(result == JOptionPane.YES_OPTION)
            writeBounds(aFile.getPath(),bounds,printBounds);

          if (aFile.exists())
            dataCenter.currentDirectory = aFile.getParentFile();  
        }
      }
    });

    JPanel panel = new JPanel(new GridLayout(1,2));
    panel.add(new JLabel("<html><font color='blue'><b>r<sub>min</sub>="+rminValue+"<br />[-d\u03A9/dx]<sub>max</sub>="+gradientValue+"</b></font></html>"));
    panel.add(printBounds);
    JFrame f = new JFrame("Splittings consistency test");
    f.setLayout(new BorderLayout());
    f.add("Center",tabbedPane);
    f.add("South",panel);
    f.setSize(700,600);
    f.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
    f.setVisible(true);    
  }

  ///////////////////////////////////////////////////////////
  // This produces a series of plots to check the consistency
  // of the rotational splittings with the rotation kernels,
  // assuming a rotation profile with a negative gradient.
  ///////////////////////////////////////////////////////////
  private void plotSplittingsConsistencyR2(double RotaSurf) {
    if (!dataCenter.modelLoaded) {
      JOptionPane.showMessageDialog(
        plotSplittingsConsistencyR2Button,
        "No model loaded, please load the model first.",
        "No model",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }

    int i, j, k, size;
    int ngrid = dataCenter.grid.length;
    double Rsurf = dataCenter.grid[ngrid-1];
    double cnst, lfreq, ufreq, dfreq;
    final double[][][] bounds = findBoundsR2(rminValue);
    ArrayList<Double> lowerValues = new ArrayList<Double>();
    ArrayList<Double> upperValues = new ArrayList<Double>();
    ArrayList<Double> freq        = new ArrayList<Double>();
    double[] x, y, error;
    Double ErrorRSi, ErrorRSj;
    JTabbedPane tabbedPane = new JTabbedPane();
    String title, IDLtitle;
    IPlot onePlot;

    for(i=0; i<dataCenter.modes.size(); i++) {
      ErrorRSi = dataCenter.modes.get(i).ErrorRS;
      if (dataCenter.modes.get(i).l.intValue() < 1) continue;
      if (dataCenter.modes.get(i).RS == null) continue;
      if (ErrorRSi == null) continue;
      if (dataCenter.modes.get(i).kernel == null) continue;
      lowerValues.clear();
      upperValues.clear();
      freq.clear();
      lfreq = Double.POSITIVE_INFINITY;
      ufreq = Double.NEGATIVE_INFINITY;
      for(j=0; j<dataCenter.modes.size(); j++) {
        ErrorRSj = dataCenter.modes.get(j).ErrorRS;
        if (Double.isNaN(bounds[0][i][j])) continue; 
        if (Double.isNaN(bounds[1][i][j])) continue; 
        if (dataCenter.modes.get(j).RS == null) continue;
        if (ErrorRSj == null) continue;
        if (dataCenter.modes.get(j).kernel == null) continue;
        if (i==j) continue;                      
        lowerValues.add(bounds[1][i][j]*(dataCenter.modes.get(j).RS-ErrorRSj)
                       /dataCenter.modes.get(j).kernel.intRotaKernelR2[ngrid-1]
                       +(1-bounds[1][i][j])*RotaSurf*Rsurf*Rsurf
                       -gradientR2Value*bounds[3][i][j]);
        upperValues.add(bounds[0][i][j]*(dataCenter.modes.get(j).RS+ErrorRSj)
                       /dataCenter.modes.get(j).kernel.intRotaKernelR2[ngrid-1]
                       +(1-bounds[0][i][j])*RotaSurf*Rsurf*Rsurf
                       +gradientR2Value*bounds[2][i][j]);
        freq.add(dataCenter.modes.get(j).FreqObs);
        if (lfreq > dataCenter.modes.get(j).FreqObs.doubleValue())
            lfreq = dataCenter.modes.get(j).FreqObs.doubleValue();
        if (ufreq < dataCenter.modes.get(j).FreqObs.doubleValue())
            ufreq = dataCenter.modes.get(j).FreqObs.doubleValue();
      }
      if (lfreq > dataCenter.modes.get(i).FreqObs.doubleValue())
          lfreq = dataCenter.modes.get(i).FreqObs.doubleValue();
      if (ufreq < dataCenter.modes.get(i).FreqObs.doubleValue())
          ufreq = dataCenter.modes.get(i).FreqObs.doubleValue();
      dfreq = (ufreq-lfreq)*0.1;
      lfreq-=dfreq;
      ufreq+=dfreq;

      // produce plot:
      title = "(n,\u2113)=("+dataCenter.modes.get(i).kernel.n
                            +","+dataCenter.modes.get(i).l.intValue()+")";
      IDLtitle = "!17(n,!12l!17)=("+dataCenter.modes.get(i).kernel.n
                               +","+dataCenter.modes.get(i).l.intValue()+")";
      onePlot = new IPlot(title,"Frequency (in \u03BCHz)","Normalised splitting");
      onePlot.setIDLTitles(IDLtitle,"!17Frequency (in !7l!17Hz)","Normalised splitting");
      size = lowerValues.size(); 

      // the original splittings
      x = new double[]{lfreq,ufreq};
      cnst = dataCenter.modes.get(i).RS/dataCenter.modes.get(i).kernel.intRotaKernelR2[ngrid-1];
      y = new double[]{cnst,cnst};
      onePlot.addDataSet(x, y, Plot.solid, Plot.noSymbol, Color.blue);
      cnst = (dataCenter.modes.get(i).RS+ErrorRSi)
           /  dataCenter.modes.get(i).kernel.intRotaKernelR2[ngrid-1];
      y = new double[]{cnst,cnst};
      onePlot.addDataSet(x, y, Plot.dotted, Plot.noSymbol, Color.blue);
      cnst = (dataCenter.modes.get(i).RS-ErrorRSi)
           /  dataCenter.modes.get(i).kernel.intRotaKernelR2[ngrid-1];
      y = new double[]{cnst,cnst};
      onePlot.addDataSet(x, y, Plot.dotted, Plot.noSymbol, Color.blue);
      cnst = RotaSurf*Rsurf*Rsurf;
      y = new double[]{cnst,cnst};
      onePlot.addDataSet(x, y, Plot.dotted, Plot.noSymbol, Color.red);
      x = new double[]{dataCenter.modes.get(i).FreqObs};
      y = new double[]{dataCenter.modes.get(i).RS/dataCenter.modes.get(i).kernel.intRotaKernelR2[ngrid-1]};
      error = new double[]{ErrorRSi/dataCenter.modes.get(i).kernel.intRotaKernelR2[ngrid-1]};
      onePlot.addDataSet(x, y, Plot.noLine, Plot.noSymbol, Color.red);
      onePlot.replaceLowerError(4,error); 
      onePlot.replaceUpperError(4,error); 
      onePlot.setVErrorValue(4,true);

      // the constraints from the other splittings
      for(j=0;j<lowerValues.size();j++) {
        if ((lowerValues.get(j)==0.0)&&(Double.isInfinite(upperValues.get(j)))) continue;
        x = new double[]{freq.get(j),freq.get(j)};
        y = new double[]{lowerValues.get(j),upperValues.get(j)};
        onePlot.addDataSet(x, y, Plot.solid, Plot.times, Color.blue);
      }
      onePlot.findRange();

      title = "("+dataCenter.modes.get(i).kernel.n
             +","+dataCenter.modes.get(i).l.intValue()+")";
      tabbedPane.add(title,onePlot);
    }

    final JButton printBounds = new JButton("Print bounds (in LaTeX format)");
    printBounds.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        // open in current directory:
        JFileChooser fc = new JFileChooser(dataCenter.currentDirectory);

        // show the filechooser
        int result = fc.showSaveDialog(printBounds);  // put parent component

        // get filename
        if(result == JFileChooser.APPROVE_OPTION) {
          File aFile = fc.getSelectedFile();

          // check to see if file exists:
          if (aFile.exists()) {
            result = JOptionPane.showConfirmDialog(printBounds,
                 "Are you sure you want to overwrite "+aFile.getName()+"?",
                 "File exists",JOptionPane.YES_NO_OPTION);
          } else {
            result = JOptionPane.YES_OPTION;
          }

          if(result == JOptionPane.YES_OPTION)
            writeBounds(aFile.getPath(),bounds,printBounds);

          if (aFile.exists())
            dataCenter.currentDirectory = aFile.getParentFile();  
        }
      }
    });

    JPanel panel = new JPanel(new GridLayout(1,2));
    panel.add(new JLabel("<html><font color='blue'><b>\u03A9<sub>surface</sub>="+Double.toString(RotaSurf)+"<br />r<sub>min</sub>="+rminValue+"<br />[d(x\u00B2\u03A9)/dx]<sub>max</sub>="+gradientR2Value+"</b></font></html>"));
    panel.add(printBounds);
    JFrame f = new JFrame("R\u00B2 Splittings consistency test");
    f.setLayout(new BorderLayout());
    f.add("Center",tabbedPane);
    f.add("South",panel);
    f.setSize(700,600);
    f.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
    f.setVisible(true);    
  }

  ///////////////////////////////////////////////////////////
  // This finds the upper and lower bounds when comparing
  // pairs of rotation kernels.
  ///////////////////////////////////////////////////////////
  // rmin = central bound used in calculating inequalities
  ///////////////////////////////////////////////////////////
  private double[][][] findBounds(double rmin) {
    int nmodes = dataCenter.modes.size();
    int ngrid = dataCenter.grid.length;
    double cnst, ratio, lowerBound, upperBound;
    double IK1, IK2, xi1, xi2, xi1_pp, xi2_pp, eta1, eta2, eta1_pp, eta2_pp, ll;
    double[] weights = null;
    int kmin, i, j, k;

    double[][][] bounds = new double[4][nmodes][nmodes];
    // bounds[0] = lower bounds over the interval [rmin, R];
    // bounds[1] = upper bounds over the interval [rmin, R];
    // bounds[2] = integral_0^rmin max(0,bounds[0]*intRotaKernel(j) - intRotaKernel(i))
    // bounds[3] = integral_0^rmin max(0,intRotaKernel(i) - bounds[1]*intRotaKernel(j))

    for(kmin=0; kmin<ngrid; kmin++) {
      if (dataCenter.grid[kmin] >= rmin) break;
    }

    if (kmin > 0) {
      double[] aux = new double[kmin+1];
      for(i=0; i<=kmin; i++) aux[i] = dataCenter.grid[i];
      weights = Numerics.findIntegrationWeights(aux);
    }

    // initialise bounds array:
    for(i=0; i<nmodes; i++) {
      for(j=0; j<nmodes; j++) {
        bounds[0][i][j] = 0.0;
        bounds[1][i][j] = Double.POSITIVE_INFINITY;
        bounds[2][i][j] = 0.0;
        bounds[3][i][j] = 0.0;
      }
    }

    for(i=0; i<nmodes; i++) {
      if (dataCenter.modes.get(i).kernel == null) continue;
      for(j=0; j<nmodes; j++) {
        if (dataCenter.modes.get(j).kernel == null) continue;
        if (i==j) {bounds[0][i][i] = bounds[1][i][i] = 1.0; continue;}
        if ((kmin == 0) && (0!=dataCenter.modes.get(i).l.compareTo(dataCenter.modes.get(j).l))) continue;
        lowerBound = Double.POSITIVE_INFINITY;
        upperBound = Double.NEGATIVE_INFINITY;
        cnst = dataCenter.modes.get(j).kernel.intRotaKernel[ngrid-1]
             / dataCenter.modes.get(i).kernel.intRotaKernel[ngrid-1];
        for(k=Math.max(1,kmin); k<ngrid; k++) {
          ratio = cnst * dataCenter.modes.get(i).kernel.intRotaKernel[k]
                       / dataCenter.modes.get(j).kernel.intRotaKernel[k];
          if (lowerBound > ratio) lowerBound = ratio;
          if (upperBound < ratio) upperBound = ratio;
        }
        // treat center separately
        if ((kmin == 0)&&(dataCenter.modes.get(i).kernel.equations != Pulsations.FromFile)
                       &&(dataCenter.modes.get(j).kernel.equations != Pulsations.FromFile)) {
          IK1     = dataCenter.modes.get(i).kernel.IK;
          IK2     = dataCenter.modes.get(j).kernel.IK;
          xi1     = dataCenter.modes.get(i).kernel.vDisp_center;
          xi2     = dataCenter.modes.get(j).kernel.vDisp_center;
          xi1_pp  = dataCenter.modes.get(i).kernel.vDisp_center_pp;
          xi2_pp  = dataCenter.modes.get(j).kernel.vDisp_center_pp;
          eta1    = dataCenter.modes.get(i).kernel.hDisp_center;
          eta2    = dataCenter.modes.get(j).kernel.hDisp_center;
          eta1_pp = dataCenter.modes.get(i).kernel.hDisp_center_pp;
          eta2_pp = dataCenter.modes.get(j).kernel.hDisp_center_pp;
          ll      = dataCenter.modes.get(i).l.doubleValue();
          if (1==dataCenter.modes.get(i).l.intValue()) {
            ratio = cnst*(IK2*(xi1_pp-eta1_pp)*(xi1_pp-eta1_pp))
                        /(IK1*(xi2_pp-eta2_pp)*(xi2_pp-eta2_pp));
          } else {
            ratio = cnst* IK2*((xi1-eta1)*(xi1-eta1)+(ll*ll+ll-2.0)*eta1*eta1)
                        /(IK1*((xi2-eta2)*(xi2-eta2)+(ll*ll+ll-2.0)*eta2*eta2));
          }
          if (lowerBound > ratio) lowerBound = ratio;
          if (upperBound < ratio) upperBound = ratio;
        }
        bounds[0][i][j] = lowerBound;
        bounds[1][i][j] = upperBound;
        if (kmin > 0) {
          for(k=0; k<=kmin; k++) {
            bounds[2][i][j] += weights[k]*Math.max(0.0,lowerBound*dataCenter.modes.get(j).kernel.intRotaKernel[k]
                                         /dataCenter.modes.get(j).kernel.intRotaKernel[ngrid-1]
                                         -dataCenter.modes.get(i).kernel.intRotaKernel[k]
                                         /dataCenter.modes.get(i).kernel.intRotaKernel[ngrid-1]);
            bounds[3][i][j] += weights[k]*Math.max(0.0,dataCenter.modes.get(i).kernel.intRotaKernel[k]
                                         /dataCenter.modes.get(i).kernel.intRotaKernel[ngrid-1]
                                         -upperBound*dataCenter.modes.get(j).kernel.intRotaKernel[k]
                                         /dataCenter.modes.get(j).kernel.intRotaKernel[ngrid-1]);
          }
        }
      }
    }
    return bounds;
  }

  ///////////////////////////////////////////////////////////
  // This produces a series of plots to check the consistency
  // of the rotational splittings with the rotation kernels,
  // assuming a rotation profile with a negative gradient.
  ///////////////////////////////////////////////////////////
  // rmin = central bound used in calculating inequalities
  ///////////////////////////////////////////////////////////
  private double[][][] findBoundsR2(double rmin) {
    int kmin, i, j, k;
    double cnst, ratio, lowerBound, upperBound;
    double IK1, IK2, xi1, xi2, xi1_pp, xi2_pp, eta1, eta2, eta1_pp, eta2_pp, ll;
    int nmodes = dataCenter.modes.size();
    int ngrid = dataCenter.grid.length;
    double[] weights = null;

    double[][][] bounds = new double[4][nmodes][nmodes];
    // bounds[0] = lower bounds over the interval [rmin, R];
    // bounds[1] = upper bounds over the interval [rmin, R];
    // bounds[2] = integral_0^rmin max(0,bounds[0]*intRotaKernelR2(j) - intRotaKernelR2(i))
    // bounds[3] = integral_0^rmin max(0,intRotaKernelR2(i) - bounds[1]*intRotaKernelR2(j))

    for(kmin=0; kmin<ngrid; kmin++) {
      if (dataCenter.grid[kmin] >= rmin) break;
    }

    if (kmin > 0) {
      double[] aux = new double[kmin+1];
      for(i=0; i<=kmin; i++) aux[i] = dataCenter.grid[i];
      weights = Numerics.findIntegrationWeights(aux);
    }

    // initialise bounds array:
    for(i=0; i<nmodes; i++) {
      for(j=0; j<nmodes; j++) {
        bounds[0][i][j] = 0.0;
        bounds[1][i][j] = Double.POSITIVE_INFINITY;
        bounds[2][i][j] = 0.0;
        bounds[3][i][j] = 0.0;
      }
    }

    for(i=0; i<nmodes; i++) {
      if (dataCenter.modes.get(i).kernel == null) continue;
      for(j=0; j<nmodes; j++) {
        if (dataCenter.modes.get(j).kernel == null) continue;
        if (i==j) {bounds[0][i][i] = bounds[1][i][i] = 1.0; continue;}
        if ((kmin==0)&&(0!=dataCenter.modes.get(i).l.compareTo(dataCenter.modes.get(j).l))) continue;
        lowerBound = Double.POSITIVE_INFINITY;
        upperBound = Double.NEGATIVE_INFINITY;
        cnst = dataCenter.modes.get(j).kernel.intRotaKernelR2[ngrid-1]
             / dataCenter.modes.get(i).kernel.intRotaKernelR2[ngrid-1];
        for(k=Math.max(1,kmin); k<ngrid; k++) {
          ratio = cnst * dataCenter.modes.get(i).kernel.intRotaKernelR2[k]
                       / dataCenter.modes.get(j).kernel.intRotaKernelR2[k];
          if (lowerBound > ratio) lowerBound = ratio;
          if (upperBound < ratio) upperBound = ratio;
        }
        // treat center separately
        if ((kmin == 0)&&(dataCenter.modes.get(i).kernel.equations != Pulsations.FromFile)
                       &&(dataCenter.modes.get(j).kernel.equations != Pulsations.FromFile)) {
          IK1     = dataCenter.modes.get(i).kernel.IK;
          IK2     = dataCenter.modes.get(j).kernel.IK;
          xi1     = dataCenter.modes.get(i).kernel.vDisp_center;
          xi2     = dataCenter.modes.get(j).kernel.vDisp_center;
          xi1_pp  = dataCenter.modes.get(i).kernel.vDisp_center_pp;
          xi2_pp  = dataCenter.modes.get(j).kernel.vDisp_center_pp;
          eta1    = dataCenter.modes.get(i).kernel.hDisp_center;
          eta2    = dataCenter.modes.get(j).kernel.hDisp_center;
          eta1_pp = dataCenter.modes.get(i).kernel.hDisp_center_pp;
          eta2_pp = dataCenter.modes.get(j).kernel.hDisp_center_pp;
          ll      = dataCenter.modes.get(i).l.doubleValue();
          if (1==dataCenter.modes.get(i).l.intValue()) {
            ratio = cnst*(IK2*(xi1_pp-eta1_pp)*(xi1_pp-eta1_pp))
                        /(IK1*(xi2_pp-eta2_pp)*(xi2_pp-eta2_pp));
          } else {
            ratio = cnst* IK2*((xi1-eta1)*(xi1-eta1)+(ll*ll+ll-2.0)*eta1*eta1)
                        /(IK1*((xi2-eta2)*(xi2-eta2)+(ll*ll+ll-2.0)*eta2*eta2));
          }
          if (lowerBound > ratio) lowerBound = ratio;
          if (upperBound < ratio) upperBound = ratio;
        }
        bounds[0][i][j] = lowerBound;
        bounds[1][i][j] = upperBound;
        if (kmin > 0) {
          for(k=0; k<=kmin; k++) {
            bounds[2][i][j] += weights[k]*Math.max(0.0,lowerBound*dataCenter.modes.get(j).kernel.intRotaKernelR2[k]
                                         /dataCenter.modes.get(j).kernel.intRotaKernelR2[ngrid-1]
                                         -dataCenter.modes.get(i).kernel.intRotaKernelR2[k]
                                         /dataCenter.modes.get(i).kernel.intRotaKernelR2[ngrid-1]);
            bounds[3][i][j] += weights[k]*Math.max(0.0,dataCenter.modes.get(i).kernel.intRotaKernelR2[k]
                                         /dataCenter.modes.get(i).kernel.intRotaKernelR2[ngrid-1]
                                         -upperBound*dataCenter.modes.get(j).kernel.intRotaKernelR2[k]
                                         /dataCenter.modes.get(j).kernel.intRotaKernelR2[ngrid-1]);
          }
        }
      }
    }
    return bounds;
  }

  ///////////////////////////////////////////////////////////
  // This writes the lower and upper bounds deduced from
  // splittings inequalities to a file.
  ///////////////////////////////////////////////////////////
  private void writeBounds(String filename, double[][][] bounds, JComponent parent) {
    int i, j, size=dataCenter.modes.size();
    try {
      PrintWriter out = new PrintWriter(new FileWriter(new File(filename)));

      out.println("%%% Lower bounds %%%");
      out.printf("\\begin{tabular}{*{%d}c}%n",size+1);
      out.printf("\\hline%n");
      out.printf("\\hline%n");
      for(j=0; j<size; j++) out.printf(" & $I_%d$",j);
      out.printf(" \\\\%n");
      out.printf("\\hline%n");
      for(i=0; i<size; i++) {
        out.printf("$I_%d$",i);
        for(j=0; j<size; j++) {
          out.printf(" & %.5f",bounds[0][i][j]);
        }
        out.printf(" \\\\%n");
      }
      out.printf("\\end{tabular}%n%n");

      out.println("%%% Upper bounds %%%");
      out.printf("\\begin{tabular}{*{%d}c}%n",size+1);
      out.printf("\\hline%n");
      out.printf("\\hline%n");
      for(j=0; j<size; j++) out.printf(" & $I_%d$",j);
      out.printf(" \\\\%n");
      out.printf("\\hline%n");
      for(i=0; i<size; i++) {
        out.printf("$I_%d$",i);
        for(j=0; j<size; j++) {
          out.printf(" & %.5f",bounds[1][i][j]);
        }
        out.printf(" \\\\%n");
      }
      out.printf("\\end{tabular}%n");
      out.close();
    } catch(Exception e) {
      System.out.println(e);
      JOptionPane.showMessageDialog(
        parent,
        "Unable to write to file \""+filename+"\"",
        "Failure",
        JOptionPane.WARNING_MESSAGE
      );
      return;
    }
  }
}
