/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008, 2009, 2011, 2014
*
****************************************************************************/
import inversion_applet.DataCenter;
import inversion_applet.FrequenciesPage;
import inversion_applet.RotationInversionPage;
import inversion_applet.StructuralInversionPage;
import inversion_applet.MeanDensityEstimatePage;
//import inversion_applet.AcousticTimePage;
//import inversion_applet.SmallSeparationPage;
import inversion_applet.KernelPage;
import inversion_applet.ModelPage;
import inversion_applet.EchellePage;
import inversion_applet.ConstantsPage;
import javax.swing.JTabbedPane;
import javax.swing.JFrame;
import javax.swing.JApplet;
import java.awt.BorderLayout;
import java.awt.Container;

/*************************************************************************
**
**    InversionKit
**                                              Version 2.1, September 2014
**
*************************************************************************/

public class InversionKit extends JApplet {

  DataCenter dataCenter;        // this contains all of the data
  RotationInversionPage rotationInversionPage;  
                                // this contains the tab with the rotation inversion results
  StructuralInversionPage structuralInversionPage;  
                                // this contains the tab with the rotation inversion results
  MeanDensityEstimatePage meanDensityEstimatePage;  
                                // this contains the tab that allows one to estimate the mean density
  /*AcousticTimePage acousticTimePage;  
                                // this contains the tab that allows one to estimate the acoustic radius
  SmallSeparationPage smallSeparationPage;  */
                                // this contains the tab that allows one to estimate the age indicator
  ModelPage modelPage;          // this contains information on the model that is loaded
  KernelPage kernelPage;        // this contains information on the kernels that are loaded
  FrequenciesPage frequenciesPage;  // this contains information on the
                                // frequency and rotational splittings used in various inversions
  EchellePage echellePage;      // this shows an echelle diagram of the observational and theoretical
                                // frequencies
  ConstantsPage constantsPage;  // this shows an echelle diagram of the observational and theoretical
                                // frequencies
 
  /*
  **  This is the entry point if the program is run as an applet.
  */
  public void init() {
    dataCenter = new DataCenter(true);
    initProgram();
  }

  /*
  **  This is the entry point if the program is run as a standalone application.
  */
  public static void main(String args[]) {
    InversionKit myApplet = new InversionKit();
    myApplet.dataCenter = new DataCenter(false);
    myApplet.initProgram();

    JFrame f = new JFrame("InversionKit 2.1");
    f.getContentPane().add("Center", myApplet);
    f.setSize(800,800);
    f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    f.setVisible(true);
  }

  /*
  **  Initialise the program
  */
  public void initProgram() {
    rotationInversionPage = new RotationInversionPage(dataCenter);
    rotationInversionPage.initProgram();

    structuralInversionPage = new StructuralInversionPage(dataCenter);
    structuralInversionPage.initProgram();

    meanDensityEstimatePage = new MeanDensityEstimatePage(dataCenter);
    meanDensityEstimatePage.initProgram();

    /*acousticTimePage = new AcousticTimePage(dataCenter);
    acousticTimePage.initProgram();

    smallSeparationPage = new SmallSeparationPage(dataCenter);
    smallSeparationPage.initProgram();*/

    modelPage = new ModelPage(dataCenter);
    modelPage.initProgram();

    frequenciesPage = new FrequenciesPage(dataCenter);
    frequenciesPage.init();

    kernelPage = new KernelPage(dataCenter);
    kernelPage.init();

    echellePage = new EchellePage(dataCenter);
    echellePage.initProgram();

    constantsPage = new ConstantsPage(dataCenter);
    constantsPage.init();

    JTabbedPane tabbedPane = new JTabbedPane();
    tabbedPane.addTab("Rotational inversion",rotationInversionPage);
    tabbedPane.addTab("Structural inversion",structuralInversionPage);
    tabbedPane.addTab("<html> <u>\u03C1</u> inversion</html>",meanDensityEstimatePage);
    /*tabbedPane.addTab("\u03C4 inversion",acousticTimePage);
    tabbedPane.addTab("\u03B4f inversion",smallSeparationPage);*/
    tabbedPane.addTab("Stellar model",modelPage);
    tabbedPane.addTab("Kernels", kernelPage);
    tabbedPane.addTab("Frequency data", frequenciesPage);
    tabbedPane.addTab("Echelle diagram", echellePage);
    tabbedPane.addTab("Physical constants", constantsPage);

    Container c = getContentPane();
    c.setLayout(new BorderLayout());
    c.add("Center", tabbedPane);
  }  
}
