/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008, 2009, 2011, 2014
*
****************************************************************************/
package inversion_applet;

import inversion_applet.DataCenter;
import inversion_applet.Utilities;
import javax.swing.JComponent;
import javax.swing.JOptionPane;
import javax.swing.JTable;
import javax.swing.JLabel;
import javax.swing.table.AbstractTableModel;
import java.io.BufferedReader;
import java.util.StringTokenizer;

public class KernelDataModel extends AbstractTableModel {

  // the data with the frequency shifts is in dataCenter;
  private DataCenter dataCenter;
 
  // These contain the parent components, which intervenes
  // when using a JPaneOption object.
  private JComponent parentComponent;

  
  // the names of the columns:
  private String[] columns = {"\u2113","n",
                              "<html>\u03BD<sub>theo</sub> (\u03BCHz)</html>",
                              "<html>\u03BD<sub>obs</sub> (\u03BCHz)</html>",
                              "<html>\u03BD<sub>var</sub>-\u03BD<sub>theo</sub> (\u03BCHz)</html>",
                              "Source"};

  // the classes of the columns:
  private Class[] columnClasses = new Class[] 
    {Integer.class, Integer.class, Double.class,
     Double.class, Double.class, Integer.class};
 
  /////////////////////////////////////////////////////////
  // a constructor:
  /////////////////////////////////////////////////////////
  protected KernelDataModel(DataCenter dataCenter,
                               JComponent parentComponent) {
    this.dataCenter = dataCenter;
    this.parentComponent = parentComponent;
  }
  
  /////////////////////////////////////////////////////////
  // the number of columns:
  /////////////////////////////////////////////////////////
  public int getColumnCount() {return 6;}
  
  /////////////////////////////////////////////////////////
  // the number of rows:
  /////////////////////////////////////////////////////////
  public int getRowCount() {return dataCenter.kernels.size();}
 
  /////////////////////////////////////////////////////////
  // return the value of a cell:
  /////////////////////////////////////////////////////////
  public Object getValueAt(int row, int col) {
    if (col == 0) return dataCenter.kernels.get(row).l;
    if (col == 1) return dataCenter.kernels.get(row).n;
    if (col == 2) return dataCenter.kernels.get(row).FreqTheo;
    if (col == 3) {
      if (dataCenter.kernels.get(row).mode == null) {
        return null;
      } else {
        return dataCenter.kernels.get(row).mode.FreqObs;
      }
    }
    if (col == 4) return dataCenter.kernels.get(row).FreqVar-dataCenter.kernels.get(row).FreqTheo;
    return dataCenter.kernels.get(row).equations;
  }

  /////////////////////////////////////////////////////////
  // return the name of a column:
  /////////////////////////////////////////////////////////
  public String getColumnName(int column) {return columns[column];}
  
  /////////////////////////////////////////////////////////
  // return the class of a column:
  /////////////////////////////////////////////////////////
  public Class getColumnClass(int c) {return columnClasses[c];}
  
  /////////////////////////////////////////////////////////
  // this makes none of the cells editable:
  /////////////////////////////////////////////////////////
  public boolean isCellEditable(int row, int col) {return false;}

  /////////////////////////////////////////////////////////
  /** set the value of a particular cell: */
  /////////////////////////////////////////////////////////
  public void setValueAt(Object obj, int row, int col) { }
  
  /////////////////////////////////////////////////////////
  /** This sorts the theoretical modes according to
   *  l and frequency. */
  /////////////////////////////////////////////////////////
  public void sort() { dataCenter.sortKernels(); }

  /////////////////////////////////////////////////////////
  /** remove a row: */
  /////////////////////////////////////////////////////////
  public void removeRow(int i) {
    if (i >= dataCenter.kernels.size()) return;
    if (dataCenter.kernels.get(i).mode != null) {
      dataCenter.kernels.get(i).mode.kernel = null;
    }
    dataCenter.kernels.remove(i);
  }
}
